"""
Utilities for use with remote processes.
"""

import sys


__all__ = [
    'OrderedDict',
    'ChangeDirectory',
]


# dictionary type that preserves insertion order if not deleting an element.
# (this is technically just an implementation detail of CPython 3.6)
if sys.version_info >= (3, 6):
    OrderedDict = dict
else:
    from collections import OrderedDict


class ChangeDirectory(object):

    """
    Context manager for temporarily changing current working directory in the
    context of the given ``os`` module.

    :param str path: new path name
    :param _os: module with ``getcwd`` and ``chdir`` functions
    """

    def __init__(self, path, _os):
        self._os = _os
        # Contrary to common implementations of a similar context manager,
        # we change the path immediately in the constructor. That enables
        # this utility to be used without any 'with' statement:
        if path:
            self._restore = _os.getcwd()
            _os.chdir(path)
        else:
            self._restore = None

    def __enter__(self):
        """Enter 'with' context."""
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Exit 'with' context and restore old path."""
        if self._restore:
            self._os.chdir(self._restore)



"""
RPC client utilities.
"""

from __future__ import absolute_import

import sys

from . import ipc


__all__ = [
    'RemoteProcessClosed',
    'RemoteProcessCrashed',
    'Client',
]


# Needed for py2 compatibility, otherwise could just use contextlib.ExitStack:
class NoLock(object):
    def __enter__(self):
        pass
    def __exit__(self, *args):
        pass


class RemoteProcessClosed(RuntimeError):
    """The MAD-X remote process has already been closed."""
    pass


class RemoteProcessCrashed(RuntimeError):
    """The MAD-X remote process has crashed."""
    pass


class Client(object):

    """
    Base class for a very lightweight synchronous RPC client.

    Uses a connection that shares the interface with :class:`Connection` to
    do synchronous RPC. Synchronous IO means that currently callbacks /
    events are impossible.
    """

    module = 'minrpc.service'

    def __init__(self, conn, lock=None, proc=None):
        """Initialize the client with a :class:`Connection` like object."""
        self._conn = conn
        self._good = True
        self._lock = lock or NoLock()
        self._proc = proc

    def __del__(self):
        """Close the client and the associated connection with it."""
        try:
            self.close()
        except (RemoteProcessCrashed, RemoteProcessClosed,
                IOError, EOFError, OSError, ValueError):
            # catch ugly follow-up warnings after a MAD-X process has crashed
            pass

    def __bool__(self):
        return self._good and not self.closed

    __nonzero__ = __bool__
    good = property(__bool__)

    @classmethod
    def spawn_subprocess(cls, lock=None, **Popen_args):
        """
        Create client for a backend service in a subprocess.

        You can use the keyword arguments to pass further arguments to
        Popen, which is useful for example, if you want to redirect STDIO
        streams.
        """
        args = [sys.executable, '-m', cls.module]
        conn, proc = ipc.spawn_subprocess(args, **Popen_args)
        return cls(conn, lock=lock, proc=proc), proc

    def close(self):
        """Close the connection gracefully, stop the remote service."""
        if self.good:
            self._conn.send(('close', ()))
        self._conn.close()
        if self._proc:
            self._proc.wait()

    @property
    def closed(self):
        """Check if connection is closed."""
        return self._conn.closed

    def _request(self, kind, *args):
        """Communicate with the remote service synchronously."""
        with self._lock:
            if self.closed:     raise RemoteProcessClosed()
            if not self._good:  raise RemoteProcessCrashed()
            try:
                response = self._communicate((kind, args))
            except (IOError, EOFError, OSError, ValueError):
                self._good = False
                self._conn.close()
                raise RemoteProcessCrashed()
        return self._dispatch(response)

    def _communicate(self, message):
        """Transmit one message and wait for the answer."""
        self._conn.send(message)
        return self._conn.recv()

    def _dispatch(self, response):
        """Dispatch an answer from the remote service."""
        kind, args = response
        handler = getattr(self, '_dispatch_%s' % (kind,))
        return handler(*args)

    def _dispatch_exception(self, exc_type, message):
        """Dispatch an exception."""
        # Raise a wrapper exception type to avoid problems if the constructor
        # expects a different arguments than a message string:
        raise type(exc_type.__name__, (exc_type,), {
            '__str__': lambda *args: message,
            '__init__': lambda *args: None})

    def _dispatch_data(self, data):
        """Dispatch returned data."""
        return data

    def get_module(self, qualname):
        """Get proxy for module in the remote process."""
        return RemoteModule(self, qualname)


class RemoteModule(object):

    """Wrapper for :mod:`cpymad.libmadx` in a remote process."""

    def __init__(self, client, module):
        """Store the client connection."""
        self.__client = client
        self.__module = module

    def __bool__(self):
        return bool(self.__client)

    __nonzero__ = __bool__

    def __getattr__(self, funcname):
        """Resolve all attribute accesses as remote function calls."""
        def DeferredMethod(*args, **kwargs):
            return self.__client._request('function_call', self.__module,
                                          funcname, args, kwargs)
        return DeferredMethod



"""
Installation script for minrpc.

Usage:
    python setup.py install
"""

from setuptools import setup
from distutils.util import convert_path


def read_file(path):
    """Read a file in binary mode."""
    with open(convert_path(path), 'rb') as f:
        return f.read()


def exec_file(path):
    """Execute a python file and return the `globals` dictionary."""
    namespace = {}
    exec(read_file(path), namespace, namespace)
    return namespace


def get_long_description():
    """Compose a long description for PyPI."""
    long_description = None
    try:
        long_description = read_file('README.rst').decode('utf-8')
        long_description += '\n' + read_file('COPYING.rst').decode('utf-8')
        long_description += '\n' + read_file('CHANGES.rst').decode('utf-8')
    except (IOError, UnicodeDecodeError):
        pass
    return long_description


def get_setup_args():
    """Accumulate metadata for setup."""
    long_description = get_long_description()
    metadata = exec_file('minrpc/__init__.py')
    return dict(
        name='minrpc',
        version=metadata['__version__'],
        description=metadata['__summary__'],
        long_description=long_description,
        author=metadata['__author__'],
        author_email=metadata['__author_email__'],
        url=metadata['__uri__'],
        license=metadata['__license__'],
        classifiers=metadata['__classifiers__'],
        packages = [
            "minrpc",
        ],
        include_package_data=True, # include files matched by MANIFEST.in
        install_requires=[
            'setuptools',
        ],
    )


def main():
    """Execute setup with parameters from ``sys.argv``."""
    setup_args = get_setup_args()
    setup(**setup_args)


if __name__ == '__main__':
    main()



# encoding: utf-8
from __future__ import unicode_literals

__title__ = 'minrpc'
__version__ = '0.0.9'

__summary__ = 'Minimalistic RPC utility (DO NOT USE!)'
__uri__ = 'https://github.com/hibtc/minrpc'

__author__ = 'Thomas Gläßle'
__author_email__ = 't_glaessle@gmx.de'

__support__ = __author_email__

__license__ = 'GPLv3+'
__copyright__ = 'Copyright 2016 HIT Betriebs GmbH'

__credits__ = """
Author and current maintainer:

    - Thomas Gläßle <t_glaessle@gmx.de>
"""

__classifiers__ = [
    'Development Status :: 4 - Beta',
    'Intended Audience :: Science/Research',
    'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
    'Operating System :: Microsoft :: Windows',
    'Operating System :: POSIX :: Linux',
    'Programming Language :: Python :: 2.7',
    'Programming Language :: Python :: 3.3',
    'Programming Language :: Python :: 3.4',
    'Programming Language :: Python :: 3.5',
    'Topic :: Software Development :: Libraries',
]


def get_copyright_notice():
    from pkg_resources import resource_string
    return resource_string('minrpc', 'COPYING.txt').decode('utf-8')
